(** Basic geometry routines *)

(* let d = 3 *)
(* let zero = [|0.0; 0.0; 0.0;|] *)

let dimension = Array.length

(** Compute the barycenter of a set of points [x] *)
let barycenter x =
  let d = dimension x.(0) in
  let n = Array.length x in
  let c = Array.make d 0.0 in
  for i=0 to n-1 do
    for j=0 to d-1 do
      c.(j) <- c.(j) +. x.(i).(j);
    done;
  done;
  for i=0 to d-1 do
    c.(i) <- c.(i) /. (float n);
  done;
  c

let dot x y = 
  let d = dimension x in
  let s = ref 0.0 in
  for i=0 to d-1 do s:= !s +. x.(i)*.y.(i); done;
  !s;;

(** Distance between two points *)
let dist x y =
  let d = dimension x in
  let r = ref 0.0 in
  for i=0 to d-1 do
    let v = x.(i) -. y.(i) in
    r := !r +. v *. v;
  done;
  sqrt !r
  
let add x y = Array.mapi (fun i xi->xi +. y.(i)) x
(*   let v = Array.make d 0.0 in *)
(*   for i=0 to d-1 do *)
(*     v.(i) <- x.(i) +. y.(i)  *)
(*   done; *)
(*   v *)

let subtract x y = Array.mapi (fun i xi->xi -. y.(i)) x
(*   let v = Array.make d 0.0 in  *)
(*   for i=0 to d-1 do *)
(*     v.(i) <- x -. y; *)
(*   done; *)
(*   v *)

let norm2 x = sqrt (dot x x)

let normalize x =
  let nrm = norm2 x in
  Array.map (fun xi-> xi /. nrm) x

(** Cross product of two vectors *)
let cross b c =
  let a = Array.make 3 0.0 in
  a.(0) <- b.(1) *. c.(2) -. b.(2) *. c.(1);
  a.(1) <- b.(2) *. c.(0) -. b.(0) *. c.(2);
  a.(2) <- b.(0) *. c.(1) -. b.(1) *. c.(0);
  a

(** Multiply a vector [x] by scalar [s] *)
let scalmult s x = Array.map (fun xi-> xi *. s) x

(** Find distance between line passing through [v0] along [v1] and
    point [p]. Return the distance and closes point on the line
    [vp]. *)
let line_point_dist (v0,v1) p =
  let r = subtract p v0 in
  let vv1 = normalize v1 in
  let t = (dot r vv1) in
  let vp = add v0 (scalmult t vv1) in
  (dist vp p),vp

(** Calculate the distance between a plane and a point [p].  [v0] is
    in the plane and [v1] is normal to the plane. Return the distance,
    the signed distance, and the closeset point in the plane. *)
let plane_point_dist (v0,v1) p =
  let r = subtract p v0 in
  let vv1 = normalize v1 in
  let sdist = (dot r vv1) in
  let dist = abs_float sdist in
  let pp = (subtract p (scalmult sdist vv1)) in
  dist, sdist, pp
    
(** Find point in set [x] which maximizes [op] which should return a
    float. *)
let max_over_points op init x =
  let n = Array.length x in
  let bestr,besti = (ref init),(ref 0) in
  for i=0 to n-1 do
    let r = op x.(i) in
    if r > ! bestr then
      (bestr := r; besti := i);
  done;
  !bestr,!besti

(** Find point in set [x] which minimizes [op] which should return a
    float. *)
let min_over_points op init x =
  let n = Array.length x in
  let bestr,besti = (ref init),(ref 0) in
  for i=0 to n-1 do
    let r = op x.(i) in
    if r < ! bestr then
      (bestr := r; besti := i);
  done;
  !bestr,!besti


(** Determine a tight bounding box for a set of points [x] which is a
    skinny matrix of points. Return the center and three principle
    axes of the bounding box. *)
let bb_princip_axes x =
  let tighten c v =					   (* Tighten axes lens *)
    let v = normalize v in
    let op p = dot v (subtract p c) in			   (* Length in pos axis dir *)
    let r1,_ = max_over_points op 0.0 x in
    let op p = dot v (subtract c p) in			   (* Length in neg axis dir *)
    let r2,_ = max_over_points op 0.0 x in		   
    let shift =  (r1 -. r2) /. 2.0 in			   (* How far to move along axis *)
    let newc = add c (scalmult shift v) in		   (* Recenter *)
    let newv = scalmult ((r1+.r2)/.2.0) v in		   (* Rescale axis *)
    newc,newv
  in
  let cent = barycenter x in				   (* Center to start *)
  (* Prim. Axis *)
  let _,idx = max_over_points (dist cent) 0.0 x in        
  let v1 = subtract x.(idx) cent in			   
  let cent,v1 = tighten cent v1 in
  (* Sec. axis *)
  let op p = let d,_ = line_point_dist (cent,v1) p in d in
  let _,idx = max_over_points op 0.0 x in		   (* Distance to prim axis *)
  let _,lp = line_point_dist (cent,v1) x.(idx) in	   (* Farthest point *)
  let v2 = subtract x.(idx) lp in			   
  let cent,v2 = tighten cent v2 in
  (* Ter. axis *)
  let v3 = cross v1 v2 in				   
  let cent,v3 = tighten cent v3 in
  (* Done *)
  cent,v1,v2,v3

(** Multipliers to produce the vertices of a bounding box given center
    and 3 prinicipal axes *)
let vert_mult = 
  [|[|  1.0;  1.0;  1.0; |];
    [|  1.0;  1.0; -1.0; |];
    [|  1.0; -1.0; -1.0; |];
    [|  1.0; -1.0;  1.0; |];
    [| -1.0;  1.0;  1.0; |];
    [| -1.0;  1.0; -1.0; |];
    [| -1.0; -1.0; -1.0; |];
    [| -1.0; -1.0;  1.0; |];|]


let bb_vertices (cent,v1,v2,v3) =
  let offset s = 
    add cent (add (scalmult s.(0) v1) (add (scalmult s.(1) v2) (scalmult s.(2) v3)))
  in
  Array.map offset vert_mult

(** Compute the dihedral angle formed between planes specified by
    [p1-p2-p3] and [p2-p3-p4]. See
    http://en.wikipedia.org/wiki/Dihedral_angles *)
let dihedral p1 p2 p3 p4 =
  let b1 = subtract p2 p1 in
  let b2 = subtract p3 p2 in
  let b3 = subtract p4 p3 in
  let b12 = cross b1 b2 in
  let b23 = cross b2 b3 in
  atan2 ((norm2 b2) *. (dot b1 b23)) (dot b12 b23)

(** Compute the angle formed by a series of 3 points *)
let angle p1 p2 p3 =
  let b1 = subtract p2 p1 in
  let b2 = subtract p3 p2 in
  (dot b1 b2) /. (norm2 b1) /. (norm2 b2)
    
(** Constant pi *)
let pi = 2.0 *. acos 0.0

let rad2deg a = a /. pi *. 180.0
