(** This file implements a PDB parser and certain operations for the
    extraction of information from it *)

open Printf;;
open Scanf;;
open Scanning;;
open String;;
open Util;;

(* Important global level module variables *)
let min_protein_length = ref 20;; (** Proteins smaller than this are treated as peptide ligands *)
let defchainsym = ref "_";;	(** Default start symbol for unlabelled chains *)
(* let elem_must_match = ref false;; (\** Atom element must match atom name to be considered for residue backbone; affects N,CA,C  *\) *)
let default_ligand = ref true;; (** Unknown atoms become ligands  *)

(** Exception when scanning goes bad; should report a line number *)
exception Read_error of string * int;;

(** List of conversions for 3 letter residue codes to 1 letter codes,
    taken from ASTRAL's website, version 1.55 Notes have this list:
    http://astral.berkeley.edu/seq.cgi?get=release-notes;ver=1.55.
    Modified conversion for GLX and ASX 6/30/2010 *)
let aa3to1 = [
  "ALA",'A'; "VAL",'V'; "PHE",'F'; "PRO",'P'; "MET",'M'; "ILE",'I'; "LEU",'L'; "ASP",'D'; "GLU",'E'; "LYS",'K';
  "ARG",'R'; "SER",'S'; "THR",'T'; "TYR",'Y'; "HIS",'H'; "CYS",'C'; "ASN",'N'; "GLN",'Q'; "TRP",'W'; "GLY",'G';
  "2AS",'D'; "3AH",'H'; "5HP",'E'; "ACL",'R'; "AIB",'A'; "ALM",'A'; "ALO",'T'; "ALY",'K'; "ARM",'R'; "ASA",'D';
  "ASB",'D'; "ASK",'D'; "ASL",'D'; "ASQ",'D'; "AYA",'A'; "BCS",'C'; "BHD",'D'; "BMT",'T'; "BNN",'A'; "BUC",'C';
  "BUG",'L'; "C5C",'C'; "C6C",'C'; "CCS",'C'; "CEA",'C'; "CHG",'A'; "CLE",'L'; "CME",'C'; "CSD",'A'; "CSO",'C';
  "CSP",'C'; "CSS",'C'; "CSW",'C'; "CXM",'M'; "CY1",'C'; "CY3",'C'; "CYG",'C'; "CYM",'C'; "CYQ",'C'; "DAH",'F';
  "DAL",'A'; "DAR",'R'; "DAS",'D'; "DCY",'C'; "DGL",'E'; "DGN",'Q'; "DHA",'A'; "DHI",'H'; "DIL",'I'; "DIV",'V';
  "DLE",'L'; "DLY",'K'; "DNP",'A'; "DPN",'F'; "DPR",'P'; "DSN",'S'; "DSP",'D'; "DTH",'T'; "DTR",'W'; "DTY",'Y';
  "DVA",'V'; "EFC",'C'; "FLA",'A'; "FME",'M'; "GGL",'E'; "GLZ",'G'; "GMA",'E'; "GSC",'G'; "HAC",'A'; "HAR",'R';
  "HIC",'H'; "HIP",'H'; "HMR",'R'; "HPQ",'F'; "HTR",'W'; "HYP",'P'; "IIL",'I'; "IYR",'Y'; "KCX",'K'; "LLP",'K';
  "LLY",'K'; "LTR",'W'; "LYM",'K'; "LYZ",'K'; "MAA",'A'; "MEN",'N'; "MHS",'H'; "MIS",'S'; "MLE",'L'; "MPQ",'G';
  "MSA",'G'; "MSE",'M'; "MVA",'V'; "NEM",'H'; "NEP",'H'; "NLE",'L'; "NLN",'L'; "NLP",'L'; "NMC",'G'; "OAS",'S';
  "OCS",'C'; "OMT",'M'; "PAQ",'Y'; "PCA",'E'; "PEC",'C'; "PHI",'F'; "PHL",'F'; "PR3",'C'; "PRR",'A'; "PTR",'Y';
  "SAC",'S'; "SAR",'G'; "SCH",'C'; "SCS",'C'; "SCY",'C'; "SEL",'S'; "SEP",'S'; "SET",'S'; "SHC",'C'; "SHR",'K';
  "SOC",'C'; "STY",'Y'; "SVA",'S'; "TIH",'A'; "TPL",'W'; "TPO",'T'; "TPQ",'A'; "TRG",'K'; "TRO",'W'; "TYB",'Y';
  "TYQ",'Y'; "TYS",'Y'; "TYY",'Y'; "AGM",'R'; "GL3",'G'; "SMC",'C'; "CGU",'E'; "CSX",'C'; 
  "GLX",'E'; "ASX",'N';
]
  (* Old codes for GLX and ASX *)
(* "GLX",'Z'; "ASX",'B'; *)


(** A hash to be used for converting the above 3-letter to 1-letter codes  *)
let aacodes =
  let h = Hashtbl.create (List.length aa3to1) in
  List.iter (fun (x,y)->Hashtbl.add h x y) aa3to1;
  h;;


(** Standard amino acids only, no extended versions *)
let aa3to1_std = [
  "ALA",'A'; "VAL",'V'; "PHE",'F'; "PRO",'P'; "MET",'M'; "ILE",'I'; "LEU",'L'; "ASP",'D'; "GLU",'E'; "LYS",'K';
  "ARG",'R'; "SER",'S'; "THR",'T'; "TYR",'Y'; "HIS",'H'; "CYS",'C'; "ASN",'N'; "GLN",'Q'; "TRP",'W'; "GLY",'G';
]

(** Standard amino acids only, no extended versions *)
let aacodes_std =
  let h = Hashtbl.create (List.length aa3to1_std) in
  List.iter (fun (x,y)->Hashtbl.add h x y) aa3to1_std;
  h;;

(** Global table to use for lookup *)
let aa_table = ref aacodes;;

(** Function that controls mapping of 3-letter codes to 1-letter codes*)
let map_aa3to1 ?(line) a3 =
  let line = match line with Some(l)->l | None -> -1 in
  let present = Hashtbl.mem !aa_table a3 in
  let msg = sprintf "Line %d, Bad residue name %s\n" line a3 in
  if present then Hashtbl.find !aa_table a3
  else Err.decide msg 'X'

(** Hash to map one letter to three letter codes *)
let aa1to3 =
  let aalist = ("UNK",'X')::aa3to1_std in
  let h = Hashtbl.create (List.length aalist) in
  List.iter (fun (x,y) -> Hashtbl.add h y x) aalist;
  h

(** Function that controls mapping of 1-letter codes to 3-letter codes*)
let map_aa1to3 a1 =
  try Hashtbl.find aa1to3 a1
  with Not_found -> failwith (sprintf "Can't find 3-letter code for residue %c" a1)

(** Convert potentially nonstandard 3-letter to standard 3-letter code *)
let map_aa3to3std a3 = map_aa1to3 (map_aa3to1 a3)

(* ** Sequence to numeric conversion routiners ** *)

let stdaaorder = "ARNDCQEGHILKMFPSTWYV";; (** Single letter to integer conversion *)
let nstdaa = String.length stdaaorder;;	(** Number of standard AAs*)

let totaaorder = "ARNDCQEGHILKMFPSTWYVBZX*";; (** Single letter to integer conversion *)
let ntotaa = String.length totaaorder;;	(** Total number of allowed AAs *)

(** Array of integers for the numeric codes of amino acids, standard only *)
let aacodes1 =
  let a = Array.make 256 (-1) in
  for i=0 to nstdaa-1 do
    a.(int_of_char stdaaorder.[i]) <- i;
  done;
  a;;

(** Convert a single letter AA to a numeric code, standard only*)
let c2i c = aacodes1.(int_of_char c);;
    
(** Hash containing the recognized nucleotide codes *)
let nuclcodes = 
  let nucl_lets = ["A";"C";"G";"T";"U";"+U"; "DA";"DC";"DG";"DT";"DU";] in
  let h = Hashtbl.create (List.length nucl_lets) in
  List.iter (fun x->Hashtbl.add h x true) nucl_lets;
  h;;

let ligcodes = Hashtbl.create 100;; (** Hash containing recognized ligands, must be filled *)
(** Add code as allowed ligand *)
let add_ligand_codes lc =
  List.iter (fun x -> Hashtbl.add ligcodes x true) lc;;

(** Process a simple flat text file to add ligands *)
let add_ligand_file ligfile = 
  let f = Scanning.from_file ligfile in
  while not (Scanning.end_of_input f) do
    Hashtbl.add ligcodes (bscanf f " %s " ident) true;
  done;;

(* (\** Check to see if a resname describes water *\) *)
(* let iswater c = c="HOH";; *)

(** Types of atoms recognized *)
type atomtype = Amino_Acid | Nucleic_Acid | Ligand | Water | Unknown;;
(** Determine type of atom based on its ResName  *)
let get_atom_type c =
  let mem = Hashtbl.mem in
  if mem aacodes c then Amino_Acid
  else if mem nuclcodes c then Nucleic_Acid
  else if c = "HOH" then Water
  else if !default_ligand then Ligand
  else if mem ligcodes c then Ligand
  else Unknown;;

(** Record for parsing information in each ATOM line of a PDB file  *)
type atom = {
  hetatom:bool;				(** Is this a HETATM *)
  serial:int;				(** Serial num of the atom *)
  namesym:string;			(** Atom symbol (C,N, etc.) *)
  namebranch:string;			(** Branch (A, B, G, etc.) *)
  altLoc:char;				(** Is this an alternate location, usually blank *)
  resName:string;			(** Name of the residue *)
  chainID:string;			(** Which logical chain is this on, defaults to '_' if not given *)
  resSeq:int;				(** Residue number for this atom *)
  icode:char;				(** Insertion code, usually for out of order residues *)
  x:float;				(** X-coordinate *)
  y:float;				(** Y-coordinate *)
  z:float;				(** Z-coordinate *)
  occupancy:float;			(** Percent occupation for this location, usually for nonblack [altLoc] *)
  tempFactor:float;			(** Temperature factor, frequently blank, defaults to 0.0 *)
  segID:string;				(** Segment ID, unknown purpos *)
  element:string;			(** Element of the atom *)
  charge:string;			(** Charge *)
  srcline:int;				(** Which line in the file the atom was taken from *)
  atomtype:atomtype;			(** Logical atom type (Amino_acid | Ligand | etc.) *)
};;

(* An empty atom for various uses *)
let empty_atom = {
  hetatom=false; serial=(-1); namesym=""; namebranch=""; altLoc=' ';
  resName="XXX"; chainID="_"; resSeq=(-1); icode=' ';
  x=0.0; y=0.0; z=0.0; occupancy=0.0; tempFactor=0.0;
  segID=""; element=""; charge=""; srcline=(-1); atomtype=Unknown;
}

(** Distance between atoms *)
let atom_dist a1 a2 =
  let dx = a1.x -. a2.x in
  let dy = a1.y -. a2.y in
  let dz = a1.z -. a2.z in
  sqrt(dx*.dx +. dy*.dy +. dz*.dz);;


(* (\** Nasty format string to read in an atom record for a pdb file *\) *)
(* let atmformati = format_of_string "%6s %5d@ %2[^\000]%2[^\000]%c %3s%_1[ ]%c %4d%c %8.3f %8.3f %8.3f %6.2f %6.2f %_6[ ]%4[^\000]%2[^\000]%2[^\000]";; *)

(** Format string for PDB ATOM records; use format_of_string to get
	the full format for printf style funcs *)
let atmformato = format_of_string "%-6s%5d %-2s%2s%c%3s %c%4d%c   %8.3f%8.3f%8.3f%6.2f%6.2f      %4s%2s%2s\n";;


(** Print and atom to the provided output channel *)
let printatom out a =
  fprintf out atmformato 
    (if a.hetatom then "HETATM" else "ATOM")
    a.serial a.namesym a.namebranch a.altLoc a.resName a.chainID.[0] a.resSeq 
    a.icode a.x a.y a.z a.occupancy a.tempFactor a.segID a.element a.charge;;

(* (\** String-only format string to read atom records for a pdb file *\) *)
(* Type generalization problems with this one *)
(* let atmformati = format_of_string *)
(*   "%6[^~]%5[^~]@ %2[^~]%2[^~]%c%3[^~]@ %c%4[^~]%c%_3[ ]%8[^~]%8[^~]%8[^~]%6[^~]%6[^~]%_6[ ]%4[^~]%2[^~]%2[^~]";; *)

(** Parse a string [s] as though it were an ATOM record and return a
    record with its contents. Empty chains are given a chainID of '_'
    by default but this may be overridden with the default argument.
*)
let parse_atom ?(defchain) ?(linenum=(-1)) ar =
  (* Pick an appopriate chain based on id and resname *)
  let assign_chain chainID resName =
    match chainID,resName with
      ' ',"HOH" -> "*"		(* Treat water specially *)
    | ' ',_ -> 
	let def = match defchain with Some(c)->c | None -> !defchainsym in
	let msg = sprintf "Line %d: Empty chain atom, default chain is %s\n" linenum def in
	Err.decide msg def
    | c,_ -> ch2str c 
  in
  (* Check field *)
  let check vars fmt default field =
    try sscanf vars fmt ident
    with _ -> 
      let msg = sprintf "Line %d: Field '%s' is empty\n" linenum field in
      Err.decide msg default
  in
  let s = 
    if String.length ar < 80 then
      let msg = sprintf "Line %d: Atom record is only %d characters wide" linenum (String.length ar) in
      Err.decide msg (sprintf "%-80s" ar)
    else ar
  in
  let sub = String.sub s in
  let resName = (sub 17 3) in
  {hetatom = (sub 0 6)="HETATM";
   serial= check (sub 6 5) " %d " 0 "serial";
   namesym=(sub 12 2);
   namebranch=(sub 14 2); 
   altLoc=s.[16]; 
   resName=resName;
   chainID = assign_chain s.[21] resName;
   resSeq= check (sub 22 4) " %d " 0 "resSeq";
   icode= s.[26];
   x=check (sub 30 8) " %f " 0.0 "x"; 
   y=check (sub 38 8) " %f " 0.0 "y"; 
   z=check (sub 46 8) " %f " 0.0 "z";
   occupancy=check (sub 54 6) " %f " 0.0 "occupancy";
   tempFactor=check (sub 60 6) " %f " 0.0 "tempFactor";
   segID=(sub 72 4);
   element=check (sub 76 2) " %s " "XX" "element";
   charge=(sub 78 2);
   srcline=linenum; 
   atomtype=get_atom_type resName;
  }

  (* let toatom hetatom serial namesym namebranch altLoc resName chainID *)
  (*     resSeq icode x y z occupancy tempFactor segID element charge = *)
  (*   let resName = sscanf resName " %s" ident in *)
  (*   {hetatom = hetatom="HETATM";  *)
  (*    serial= check serial " %d " 0 "serial";  *)
  (*    namesym=namesym; namebranch=namebranch; altLoc=altLoc; resName=resName; *)
  (*    chainID = assign_chain chainID resName; *)
  (*    resSeq= check resSeq " %d " 0 "resSeq";  *)
  (*    icode=icode;  *)
  (*    x=check x " %f " 0.0 "x"; y=check y " %f " 0.0 "y"; z=check z " %f " 0.0 "z";  *)
  (*    occupancy=check occupancy " %f " 0.0 "occupancy"; *)
  (*    tempFactor=check tempFactor " %f " 0.0 "tempFactor";  *)
  (*    segID=segID;  *)
  (*    element= *)
  (* 	begin *)
  (* 	  if element<>"" then element  *)
  (* 	  else *)
  (* 	    let msg = sprintf "Line %d: Field '%s' is empty\n" linenum "element" in *)
  (* 	    Err.decide msg "XX" *)
  (* 	end; *)
  (*    charge=charge; *)
  (*    srcline=linenum; atomtype=get_atom_type resName;} in *)
  (* (\* parse and return *\) *)
  (* let atmformati = format_of_string  *)
  (*   "%6[^~]%5[^~]@ %2[^~]%2[^~]%c%3[^~]@ %c%4[^~]%c%_3[ ]%8[^~]%8[^~]%8[^~]%6[^~]%6[^~]%_6[ ]%4[^~]%2[^~]%2[^~]" in *)
  (* sscanf ar atmformati toatom;; *)

(** Holds information on an individual residue. N, CA, and C atoms are
    given special fields which may or not be filled. With below
    parsing routine, N, CA, and C appear in the list of all atoms,
    resatoms, as well as the rest of the atoms associated with this
    residue. *)

type resrec = {
  resnum:int;
  incode:char;				(** Insertion code for residue *)
  chain:string;
  mutable n:atom option;
  mutable ca:atom option;
  mutable c:atom option;
  mutable resatoms:atom list;
}

(** Empty residue for initialization *)
let empty_res = {
  resnum=(-999); incode=' '; chain=(!defchainsym); n=None; ca=None; c=None; resatoms= [];
}

(** Type abstraction for backbone atoms  *)
type resatom = AlphaC | NitroBB | CarbBB | SCother;;

(** Determine backbone types *)
let get_resatom_type atom =
  match atom.namesym,atom.namebranch with
    " C","A " -> AlphaC
  | " C","  " -> CarbBB
  | " N","  " -> NitroBB
  | _ -> SCother;;

(** Read all atoms from file [fname] and return them as a list *)
let read_all_atoms fname =
  let b = Pervasives.open_in fname in
  let al = Queue.create () in
  let ter = ref false in
  let srcline = ref 0 in
  begin try while not !ter do
    let line = input_line b in
    incr srcline;
    let rtype = sscanf line " %6s " ident in 
    begin match rtype with
      "ATOM" | "HETATM" ->
	let atom = try 
	  parse_atom ~linenum:!srcline line 
	with  _ -> raise (Read_error (fname,!srcline)) 
	in  
	Queue.add atom al;
    | "TER" -> ();
    | "ENDMDL" -> ter := true
    | _ -> ();
    end;
  done with End_of_file ->() end; 
  Pervasives.close_in b;
  q2list al;;

(** Generic routine that groups sequential items. [pred] takes two
    items and produces true or false if they are in the same
    group. This is handed the last item in the group to be
    added. Returns a list of item lists *)
let seq_group pred al =
  let groups = Queue.create () in
  let rec iter a agrp l = match a,l with
      a,b::t -> 
	if pred a b 
	then (Queue.add b agrp; iter b agrp t)
	else begin 
	  Queue.add (q2list agrp) groups; 
	  let q = Queue.create () in
	  Queue.add b q;
	  iter b q t
	end
    | _,[] -> 
	Queue.add (q2list agrp) groups in
  begin match al with
    [] -> ();
  | h::t -> 
      let q = Queue.create () in
      Queue.add h q;
      iter h q t
  end;
  q2list groups;;


(** Create a residue from a group of atoms *)
let res_of_atoms g = 
  let rec add_atoms r al = match al with
      [] -> r
    | h::t -> 
	begin match get_resatom_type h with
	  AlphaC -> if r.ca = None then r.ca <- Some(h);
	| CarbBB -> if r.c = None then r.c <- Some(h);
	| NitroBB -> if r.n = None then r.n <- Some(h);
	| SCother -> ();
	end; 
	add_atoms r t in
  match g with
    [] -> empty_res
  | a::_ -> 
      add_atoms {resnum=a.resSeq; incode=a.icode; chain=a.chainID; n=None; 
		 ca=None;c=None; resatoms=g} g;;

(** Determines if atoms are in the same group *)
let same_atom_group a b = 
  a.resName=b.resName && a.resSeq=b.resSeq && a.chainID=b.chainID && a.icode=b.icode;;

(** Estimate of CA-CA distance *)
let ca_dist_mean = 3.8;;
(** Tolerance for deviation from mean CA-CA-distance *)
let ca_dist_tol = 1.0;;

(** Alpha distance is within a certain distance then residues are
    adjacent *)
let adjacent_alpha r s =
  match r.ca,s.ca with
    Some(ca1),Some(ca2) -> 
      let d = atom_dist ca1 ca2 in
      abs_float (d -. ca_dist_mean) < ca_dist_tol
  | _ -> false;;

(** Estimate of C-N distance *)
let cn_dist_mean = 1.33;;
(** Tolerance for deviation from mean CN-distance *)
let cn_dist_tol = 0.3;;

(** C-N distance is within a certain distance then residues are
    adjacent *)
let adjacent_cn r s =
  match r.c,s.n with
    Some(c),Some(n) -> 
      let d = atom_dist c n in
      abs_float (d -. cn_dist_mean) < cn_dist_tol
  | _ -> false;;

(** Residue indices are within 1 indicates adjacency *)
let adjacent_idx r s = abs (s.resnum - r.resnum) = 1;;

let same_atomtype r s =
  match r.ca,s.ca with
    Some(ca1),Some(ca2) -> 
      ca1.atomtype = ca2.atomtype
  | _ -> false;;

(** Tolerant criteria for residue adjacency *)
let tolerant_adjacency r s =
  let crit = [adjacent_alpha r s; adjacent_cn r s; adjacent_idx r s] in
  2 <= (List.fold_left (fun s b-> if b then s+1 else s) 0 crit)

(** Strict criteria for adjacency *)
let strict_adjacency r s =
  adjacent_alpha r s && adjacent_cn r s && adjacent_idx r s;;

(** Adjacent if chain ID letters match *)
let chain_adjacency r s = r.chain = s.chain;;

(** Whole pdb will be one protein chain *)
let always_adjacent _ _ = true;;

let really_strict_adj r s =
     adjacent_alpha r s 
  && adjacent_cn r s 
  && adjacent_idx r s
  && same_atomtype r s;;

(** Force a full backbone for good residues  *)
let fullbackbone r = match r.n,r.ca,r.c with
  Some(_),Some(_),Some(_) -> true (* ca.atomtype = Amino_Acid *)
  | _ -> false;;

(** Only require alpha carbon to be present *)
let alphaonly r = match r.ca with
    Some(_) -> true (* ca.atomtype = Amino_Acid *)
  | _ -> false;;

(** Extract all protein chains from the atom list list [al], return the
    list of residues and the remainder of the atom list. *)
let extract_residues goodres all =
  let resl = Queue.create () in
  let rest = Queue.create () in
  let rec iter l = match l with
      [] -> ()
    | g::t ->
	let r = res_of_atoms g in
	if goodres r then Queue.add r resl
	else Queue.add g rest;
	iter t in
  iter all;
  (q2list resl),(q2list rest);;

(** Type for a protein chain, flattened coordinates for easy access *)
type protein_chain = {
  chainlen:int;
  chainid:string;
  alphax:float array;
  alphay:float array;
  alphaz:float array;
  alpha_coords:float array array;	(** X,Y,Z coords as a row-based matrix *)
  residues1:string;
  residues3:string array;
  ids:int array;
  resrec:resrec array;
  alphas:atom array;
  phi:float array;			(* Array of phi angles in degrees *)
  psi:float array;			(* Array of psi angles in degrees *)
  atoms:atom array array;
}

let empty_chain = {
  chainlen=0;
  chainid="";
  alphax=[||];
  alphay=[||];
  alphaz=[||];
  alpha_coords=[||];
  residues1="";
  residues3=[||];
  ids=[||];
  resrec=[||];
  alphas=[||];
  phi=[||];
  psi=[||];
  atoms=[||];
};;

(** Type for chains of nucleic acids *)
type nuclacid_chain = {
  nachainID:string;
  acids:atom array array;
  naallatoms:atom array;
}

(** Compute phi angle between two residues, return in degrees *)
let get_phi res1 res2 =
  match res1,res2 with
    {c=Some(c1);_},{n=Some(n2);ca=Some(ca2);c=Some(c2);_} ->
      let (p1,p2,p3,p4) = ([|c1.x;c1.y;c1.z|], [|n2.x;n2.y;n2.z|],
			   [|ca2.x;ca2.y;ca2.z|], [|c2.x;c2.y;c2.z|]) in
      Geom.rad2deg (Geom.dihedral p1 p2 p3 p4)
  | _,_ -> Err.decide "Missing atom in get_phi" 0.0

(** Compute psi angle between two residues, return in degrees *)
let get_psi res1 res2 =
  match res1,res2 with
    {n=Some(n1);ca=Some(ca1);c=Some(c1);_},{n=Some(n2);_} ->
      let (p1,p2,p3,p4) = ([|n1.x;n1.y;n1.z|], [|ca1.x;ca1.y;ca1.z|], 
			   [|c1.x;c1.y;c1.z|], [|n2.x;n2.y;n2.z|]) in
      Geom.rad2deg (Geom.dihedral p1 p2 p3 p4)
  | _,_ -> Err.decide "Missing atom in get_psi" 0.0
		
(** Create arrays of phi and psi angles *)
let calc_phipsi resrecs =
  let n = Array.length resrecs in
  let phis = Array.make n nan in
  for i=1 to n-1 do
    phis.(i) <- get_phi resrecs.(i-1) resrecs.(i)
  done;
  let psis = Array.make n nan in
  for i=0 to n-2 do
    psis.(i) <- get_psi resrecs.(i) resrecs.(i+1)
  done;
  (phis,psis)

(** Convert a list of [resrec]s to a flattened represenetation, a
    [protein_chain] *)
let residuelist2chain ?chainnum l =
  let resrec = Array.of_list l in
  let chainlen = Array.length resrec in
  let alphas = Array.map 
    (fun r->match r.ca with Some(a)->a |None -> failwith "No alpha carbon") resrec in
  let chainid = match chainnum with
      None -> alphas.(0).chainID
    | Some(n) ->  alphas.(0).chainID ^ (string_of_int n) in
  let x = Array.map (fun a->a.x) alphas in
  let y = Array.map (fun a->a.y) alphas in
  let z = Array.map (fun a->a.z) alphas in
  let residues3 = Array.map (fun a->a.resName) alphas in
  let residues1 = String.make chainlen '-' in
  for i=0 to chainlen-1 do
    residues1.[i] <- map_aa3to1 ~line:alphas.(i).srcline residues3.(i)
  done;
  let ids = Array.map (fun r->r.resnum) resrec in
  let atoms = Array.map (fun r->Array.of_list r.resatoms) resrec in
  let phi,psi = calc_phipsi resrec in
  {chainlen=chainlen; chainid=chainid; alphax=x;alphay=y;alphaz=z; 
   alpha_coords=[|x;y;z;|];
   residues1=residues1; residues3=residues3; ids=ids; resrec=resrec;
   alphas=alphas; phi=phi; psi=psi; atoms=atoms;};;

(** Represents a parsed PDB file *)
type pdb = {
  proteins:protein_chain array;		(** All chains of amino acids *)
  nuclacids:nuclacid_chain array;	(** Chains of nucleic acids *)
  ligands:atom array array;		(** Ligands grouped together *)
  water:atom array;			(** Water molecules *)
  others:atom array array;		(** Other atoms, not identified *)
  unbroken_chains:protein_chain array;	(** Unbroken chains *)
  broken_chains:protein_chain array array; (** Broken chains grouped together *)
};;
 
(** Eliminate alternate conformations in a residue if present. The
    default conformation is taken from CA but blank altLoc fields are
    also accepted. Using CA for the default ensures the residue isn't
    too badly broken by eliminating some atoms but it may still
    happen. *)
let eliminate_alt_conf res =
  let c = match res.ca with 
      Some(ca)-> 
	if ca.altLoc <> ' ' then ca.altLoc (* First try to take the CA altLoc *)
	else List.fold_left 		(* O/w take first non ' ' altLoc *)
	  (fun c a->if c<>' ' then c else a.altLoc) ' ' res.resatoms
    | None-> 
	let msg = sprintf "Trying to remove altLocs for res %d chain %s with no CA\n" 
	  res.resnum res.chain in
	Err.decide msg ' '
  in
  let newatoms = List.filter (fun a->a.altLoc=c || a.altLoc=' ') res.resatoms in
  res.resatoms <- newatoms;;

(** Do some sanity checks, if [fail] is true, bail out if the pdb is
    'insane'. Right now this just checks for matching chain IDs for
    the whole chain and each residue. *)
let sanity_checks ?(fail=false) pdb =
  let sane = ref true in
  let chain_match p =
    let cm = ref true in
    for i=0 to p.chainlen-1 do
      let rc = p.resrec.(i).chain in
      let m = (rc = p.chainid) in
      if not m then 
	let msg =sprintf "Chain %s, residue %d has nonmatching chain ID %s" 
	  p.chainid p.ids.(i) rc in
	Err.warn msg;
      cm := !cm && m;
    done; !cm
  in

  for i=0 to (Array.length pdb.proteins)-1 do
    let p = pdb.proteins.(i) in
    sane := !sane && (chain_match p);
  done;
  if fail && not !sane then failwith "Insane PDB file";;

      
(** Logic to determine broken chain relationships. Chains are
    identified by the chainID of their first atom. [proteins] are a
    list of residue lists. Return [single_ch], a list of residue lists
    which are single chains, and [broken_ch], a list of residue list
    list. The first level of the list is each associated chain, second
    level the broken pieces of the chain, and third the actual
    residues belonging to that piece. If [unbreak] is true, then make
    all broken chains single chains. *)
let broken_chain_analysis unbreak proteins =
  let chains = Hashtbl.create 20 in	(* Collect the residues chain *)
  let ids = Hashtbl.create 20 in	(* Collect chain ids *)
  let iter rl = 
    match rl with
      {ca=Some(a); _}::_ -> 
	let id = a.chainID in
	Hashtbl.add chains id rl;	(* Remember this chain with ID *)
	Hashtbl.replace ids id () 	(* Remember this ID *)
    | _ -> () in
  List.iter iter proteins;
  let gather id () l =			(* Create a list of residues on the chain *)
    (List.rev (Hashtbl.find_all chains id))::l in
  let chain_groups = List.rev (Hashtbl.fold gather ids []) in
  let chain_groups =			(* 2D list of reslists *)
    if unbreak then
      List.map (fun l-> [List.concat l]) chain_groups
    else chain_groups
  in
  let single, broken = List.partition (fun l-> List.length l = 1) chain_groups in
  let single = List.map (fun l->List.hd l) single in
  single, broken;;

(** Figure out how to group chains. Apply residue adjacency according
    to [adj_chainf]. Check for broken chains using
    [broken_chain_analysis] and potentially merge them if [unbreak] is
    true. Partition out chains shorter than [min_chain_len]. Return an
    array of all chains, an array of unbroken chains, a 2D array of
    broken chains, and a list grouped atoms which didn't fit in the
    chain. *)
let protein_chain_analysis min_chain_len unbreak adj_chainf residue_lists =
  let chains = seq_group adj_chainf residue_lists in
  let short_ch, long_ch =
    List.partition (fun l->List.length l < min_chain_len) chains in
  let single_ch, broken_ch = broken_chain_analysis unbreak long_ch in
  let short_chain_groups =	  (* Short chains, probably ligands *)
    let chainatoms c = List.flatten (List.map (fun r->r.resatoms) c) in
    List.map chainatoms short_ch in
  let unbroken_chains = Array.of_list (List.map residuelist2chain single_ch) in
  let broken_chains_array_list =
    let parts_list2array pl =
      let i = ref 0 in
      let inc_reslist2chain rl = 
	let c = residuelist2chain ~chainnum:!i rl in
	incr i; c in
      Array.of_list (List.map inc_reslist2chain pl)
    in
    List.map parts_list2array broken_ch
  in
  let broken_chains = Array.of_list broken_chains_array_list in
  let all_chains = Array.append unbroken_chains (Array.concat broken_chains_array_list) in
  all_chains, unbroken_chains, broken_chains, short_chain_groups;;


(** Group nucleic acids into chains *)
let group_nucl_acids naq =
  let chains = Queue_hash.create 3 in
  let iter na =
    let chain = na.(0).chainID in
    Queue_hash.add chains chain na
  in
  Queue.iter iter naq;
  let make (chain,atoms) = 
    {nachainID=chain; acids=atoms; 
     naallatoms=Array.concat (Array.to_list atoms)} in
  Array.map make (Queue_hash.exporta chains)


(** Group the nonprotein entities according to their type. Return
    arrays of these entities. *)
let assign_nonchain_entities atom_groups_list =
  let nuclacids, ligands, others, water = 
    let newQ = Queue.create in
    newQ (), newQ (), newQ (), newQ () in
  let iter g = 
    assert (g <> []);
    match (List.hd g).atomtype with	(* Right now, only look first atom type in the list *)
      Amino_Acid ->   Queue.add (Array.of_list g) ligands
    | Nucleic_Acid -> Queue.add (Array.of_list g) nuclacids 
    | Ligand ->       Queue.add (Array.of_list g) ligands 
    | Water ->        List.iter (fun w-> Queue.add w water) g
    | Unknown ->      Queue.add (Array.of_list g) others 
  in
  List.iter iter atom_groups_list;
  let nuclacids = group_nucl_acids nuclacids in
  let q2a = Util.q2array in
  nuclacids, q2a ligands, q2a others, q2a water;;

(** Function which determines acceptable residues *)
let goodres = ref alphaonly
(** Function which determines adjacenct residues *)
let adj_chainf = ref tolerant_adjacency
(** Minimum number of atoms to bec considered a ligand *)
let min_lig_atoms = ref 8
(** Maximum number of atoms to bec considered a ligand *)
let max_lig_atoms = ref 500
(** Minimum residue length to be considered a protein chain *)
let min_chain_len = ref 10
(** Eliminate alternate conformations of residues *)
let no_alt = ref false
(** Merge potentially broken chains *)
let unbreak = ref false

(** Creates atom groups, residue groups, etc, returns pdb *)
let atoms2pdb atoms =
  let groups = seq_group same_atom_group atoms in
  let residues, nonresidues = extract_residues !goodres groups in 
  if !no_alt then List.iter eliminate_alt_conf residues;
  (* Deal with protein chains *)
  let all_ch, unbroken_ch, broken_ch, non_chain_atoms = 
    protein_chain_analysis !min_chain_len !unbreak !adj_chainf residues in
  (* Deal with other atoms -> ligands, nucleic acids, water *)
  let nuclacids, ligands, others, water =
    assign_nonchain_entities (List.append nonresidues non_chain_atoms) in
  let pdb = {
    proteins = all_ch;
    nuclacids = nuclacids;
    ligands = ligands;
    others = others;
    water = water;
    unbroken_chains = unbroken_ch;
    broken_chains = broken_ch;
  } in 
  begin match !Err.lvl with
    Err.NoCheck -> ()
  | Err.Complain -> sanity_checks pdb
  | Err.Panic -> sanity_checks ~fail:true pdb
  end;
  pdb;;

(** Main routine, read a pdb file [fname] and return a [pdb] data
    structure *)
let read_pdb fname =
  let atoms = read_all_atoms fname in
  atoms2pdb atoms;;

(* Discarded effort to deal with atom specs/coordinates in separate
    files for the floudas-hr decoy set *)
(** Read pdb from [pdbname] but replace the XYZ coordinates with the
    coordinates found in file [coordname]. The file [coordname] should
    have a header line which is skipped and otherwise order the atoms
    identically to [pdbname]. *)
let read_pdb_altcoords pdbname coordname =
  let atoms = read_all_atoms pdbname in
  let f = Pervasives.open_in coordname in
  ignore(input_line f);
  let srcline = ref 1 in
  (* Alter atoms *)
  let alter a = 
    try
      incr srcline;
      let line = input_line f in
      sscanf line " %f %f %f" (fun x y z -> {a with x=x; y=y; z=z;})
    with  _ -> raise (Read_error (coordname,!srcline)) 
  in  
  let newatoms = List.map alter atoms in
  atoms2pdb newatoms;;
  

  (** Take a slice of the residues *)
let chain_slice chain start len =
   let alphax=Array.sub chain.alphax start len in
   let alphay=Array.sub chain.alphay start len in
   let alphaz=Array.sub chain.alphaz start len in
   let alpha_coords = [|alphax;alphay;alphaz|] in
  {chainlen=len;
   chainid=chain.chainid;
   alphax=alphax;
   alphay=alphay;
   alphaz=alphaz;
   alpha_coords=alpha_coords;
   residues3=Array.sub chain.residues3 start len;
   residues1=String.sub chain.residues1 start len;
   ids=Array.sub chain.ids start len;
   resrec=Array.sub chain.resrec start len;
   alphas=Array.sub chain.alphas start len;
   phi=Array.sub chain.phi start len;
   psi=Array.sub chain.psi start len;
   atoms=Array.sub chain.atoms start len;
  };;

(** Take a slice of the residues starting with [start] and ending with [stop] *)
let chain_slice_idx chain start stop =
  let len = stop-start+1 in
  chain_slice chain start len;;

(** Return a slice starting with alpha [id1] and ending with [id2] *)
let chain_slice_aid chain id1 id2 =
  let start = ref 0 in
  while chain.ids.(!start) <> id1 do incr start done;
  let stop = ref (!start + 1) in
  while chain.ids.(!stop) <> id2 do incr stop done;
  chain_slice chain !start (!stop - !start +1);;


(* ******************************************************************************** *)

(** Fragment data structure *)
type fragment = {
    length:int;
    coords:float array array;
  }

(** Read a flat text file containing multiple fragments. Header of
    file describes number and lenght of fragments *)

let read_fragments fname = 
  let b = Scanning.from_file fname in
  let (nfrags,flen) = bscanf b "# %d %d " (fun i j->(i,j)) in
  let frags = Array.make nfrags {length=0; coords=Array.make_matrix 3 0 0.0} in
  for i=0 to nfrags-1 do 
    frags.(i) <- {length=flen;  coords=Array.make_matrix 3 flen 0.0};
    for j=0 to flen-1 do
      let (x,y,z) = bscanf b " %f %f %f " (fun x y z->(x,y,z)) in
      frags.(i).coords.(0).(j) <- x;
      frags.(i).coords.(1).(j) <- y;
      frags.(i).coords.(2).(j) <- z;
    done
  done;
  frags;;

open Filename;;
(** Return the base filename of a pdbfile *)
let file_id fname = 
  String.sub (chop_extension (basename fname)) 0 4;;

let options = CKArg.([
  "PDB Reader Options","","",Header;
  "-fullbb","","Require N, CA, C for a residue to be good",Unit(fun ()->goodres:=fullbackbone);
  "-adj","key","Control how adjacent residues are determined which in turn determines logical \
                grouping of residues into chains",(
    SymbolArb(
      true,[
	"tolerant","Check for 2/3 of CAs close, C-N close, and residue index close",(fun ()->adj_chainf:=tolerant_adjacency);
	"strict"  ,"Require CA-CA, C-N, and res index distances to be correct for chain adjacency",(fun ()->adj_chainf:=strict_adjacency);
	"chain"	  ,"Only check that the chain letter identifier is the same",(fun ()->adj_chainf:=chain_adjacency);
	"always"  ,"Turn off adjacency checking for residues",(fun ()->adj_chainf:=always_adjacent);
      ]));
  "-min_chain_len","int","Minimum length of a chain before is classified as a ligand",Set_int(min_chain_len);
  "-unbreak","","If broken chains are found, merge them",Set(unbreak);
  "-min_lig_atoms","int","Number of atoms a ligand must have to be considered in distance calculation",Set_int(min_lig_atoms);
  "-max_lig_atoms","int","Discard ligands with more atoms in distance calculation",Set_int(max_lig_atoms);
  "-stdaa","","Translate all nonstandard AA's to X for sequence",Unit(fun ()->aa_table := aacodes_std);
  "-noalt","","Do not include alternate conformations in the pdb",Set(no_alt);
])
