(** This module implements a number of command line functions on PDB
    files such as printing info, splitting into chains, extracting
    sequence, and calculating ligand distances. *)

open PDBRead;;
open Util;;

open Printf;;
open Scanf;;
open Arg;;

(* Some filters for computing distances between residues and to
   ligands *)

(** Never ignore either atom *)
let always_filt (_:atom) (_:atom) = true

(** Filter for backbone atoms *)
let bbfilt a =
  match a.namesym.[1],a.namebranch.[0] with
    'C','A' | 'N',' ' | 'C',' ' | 'O',' ' -> true
  | _ -> false;;

(** Not a hydrogen atom  *)
let noth_filt a = (a.element.[1] <> 'H');; (** Not hydrogen filter, may be broken *)

(** Primitive side chain atom check. This makes sure that the OXT atom
    that appears on some ending residues doesn't appear as part of the
    sidechain. *)
let scfilt a =
  let sym = a.namesym in let brc = a.namebranch in
  (not (bbfilt a)) 
  && (noth_filt a) 
  && (not (sym.[1]='O' && (brc.[0]='X' && brc.[1]='T')));;

(** Main chain filter usable by [protein_ligand_dists *)
let mc_filt a (_:atom) = bbfilt a

(** Side chain filter usable by [protein_ligand_dists] *)
let sc_filt a (_:atom) = scfilt a

let mode = ref ""
let flist = ref []
let ext = ref ""
let chop = ref false
let include_ligs = ref false
let chainlist = ref ""
let print_first_resnum = ref false
let fillseq = ref false
let ligfile = ref ""
let filelist = ref ""
let dist_filt = ref always_filt
let contact_dists = ref ""
let separate_on_chain = ref true
let renumber = ref false
let rechain = ref false
let nohet = ref false
let bbdebug = ref false

let print_ter b = fprintf b "%-80s\n" "TER"
let print_end b = fprintf b "%-80s\n" "END"
  
(** File name is code *)
let filecode fname = Filename.chop_extension (Filename.basename fname);;

let getcode = ref filecode;;

open CKArg

let options = [
  "PDB Tool Options","","",Header;
  "-mode","key","Global Operation mode",
  Symbol(true,(fun s->mode:=s),[
	   "info","Print information on the pdb file";
	   "sum","Print a one line summary of the structure";
	   "reform","Spit the PDB back out with possible modifications (e.g. remove ligands, renumber, etc)";
	   "seq","Extract sequences for chains in the PDB";
	   "angles","Print dihedral angles";
	   "ligdist","Calculate the distance between residue and ligand atoms";
	   "extract","Extract chains from the pdb into their own pdb file";
	   "hierarch","Print a verbose output of how pdbtool is interpreting the pdb file";
	   "pybb","Create a bounding box file usable by pymol around each chain";
	 ]);	    
  "-ext","extension","Extension for output files, default by mode (seq->.fa , ligdist->.dist) ",Set_string(ext);
  "-filelist","file","File containing list of pdb files to process",Set_string(filelist);
  "-chop","","Chop sequence lines at 80 characters (true FASTA format)",Set(chop);
  "-include_ligs","","Include ligands when extracting/splitting",Set(include_ligs);
  "-chainlist","strings","Space separated list of chains to process (default all chains). Only honoroed for mode=extract",Set_string(chainlist);
  "-renumber","","Renumber residues from 1",Set(renumber);
  "-rechain","","Reassign chains to match grouping determined by pdbread",Set(rechain);
  "-nohet","","During extract/split, print everything as an atom (no hetero atoms)",Set(nohet);
  "-ligands","file","If specified, only residues with these codes are considered ligands",Set_string(ligfile);
  "-id4","","Use output file ids which are 4-letter codes",Unit(fun ()->getcode := PDBRead.file_id);
  "-print_first_resnum",""," Print the first residue number in fasta files",Set(print_first_resnum);
  "-fillseq","","Fill in gaps in sequence with X, inhibits chain checking for extracting sequences",Set(fillseq);
  "-nosep","","Do not put sequences in separate chain-specific files",Clear(separate_on_chain);
  "-dist_filt","key","Choose which type of atom in residues to use for ligand \
                      distance computation",(
    SymbolArb(
      true,[
	"all", "Consider all residue atoms",    (fun ()->dist_filt:= always_filt);
	"mc",  "Consider only main chain atoms",(fun ()->dist_filt:= mc_filt);
	"sc",  "Consider only side chain atoms",(fun ()->dist_filt:= sc_filt);
      ]));
  "-contact_dists","'d1 [d2 ..]'","Space separated list of distances which indicate protein \
      ligand contacts, adds contact info to info printout",Set_string(contact_dists);
  "-errlvl","key","Set the error checking level",(
    Symbol(true,Err.setlvl,Err.docs));
] @ PDBRead.options

(*   "-partialbb",Unit(fun ()->goodres:=PDBRead.alphaonly)," Only require alpha carbons to be present for inclusion in sequence"; *)
(*   "-matchelem",Set(PDBRead.elem_must_match)," Element and name of an atom must match to be a residue backbone atom (CA,C->C,N->N)"; *)

let deffn s = flist := s::!flist;;
let usg = sprintf "usage: %s [options] pdbfile [pdbfile2 ...]" (Filename.basename Sys.executable_name);;

(** Fill in gaps in [seq] according to [idxs] with 'X' for unkown residues *)
let fillin_seq seq idxs =
  let n = Array.length idxs in
(*   let pos = Array.make n 0 in *)
(*   (\* Try to determine 'real' indexes for residues based on idxs. Looks *)
(*      ahead a little to see if single residues are one-off *)
(*      differences *\) *)
(*   pos.(0) <- 0; *)
(*   let loc = ref 1 in *)
(*   for i=1 to n-1 do *)
(*     if idx.(i)-idx.(i-1) = 1 then *)
(*       (pos.(i) <- !loc; incr loc) *)
(*     else if i+1<n && idx.(i+1) - idx.(i-1) = 1 then *)
(*       (p *)
(*   done; *)
  let start,stop = idxs.(0),idxs.(n-1) in
  let len = stop-start+1 in
  let fillseq = String.make len 'X' in
  for i=0 to n-1 do
    fillseq.[idxs.(i)-start] <- seq.[i]
  done;
  fillseq;;

(** Extract sequences from each protein entity into their own file *)
let pdb2seq fname pdb =
  let id = !getcode fname in
  let ext = if !ext = "" then ".fa" else !ext in
  let fn chain =
    let c = chain.chainid in
    let f = 
      if !separate_on_chain then
	open_out (sprintf "%s%s%s" id c ext) 
      else
	open_out_gen [Open_append; Open_creat] 0o640 (sprintf "%s%s" id ext)
    in
    if !print_first_resnum then
      fprintf f "> %s%s %d\n" id c chain.ids.(0)
    else fprintf f "> %s%s\n" id c;
    let seq = 
      if !fillseq then 
	fillin_seq chain.residues1 chain.ids
      else chain.residues1 in
    let n = String.length seq in
    let nc = ref 0 in
    for i=0 to n-1 do
      if !chop && !nc>0 && !nc mod 80 = 0 then fprintf f "\n"; 
      fprintf f "%c" seq.[i];
      if not !fillseq && chain.alphas.(i).chainID <> c then 
	eprintf "Warning: chain %s, res %d has chainid %s\n" c i chain.alphas.(i).chainID;
      incr nc;
    done;
    fprintf f "\n";
    close_out f in
  Array.iter fn pdb.proteins;;

(** Calculate the distances between the protein residues in [prots]
    and the ligand atom arrays in [ligs]. Returns an array of arrays
    of the minimum distance of each residue atom to any ligand atom, a
    row for each chain, a column for each residue in the
    chain. Contents are distance*atom pairs where the atom is the
    ligand atom closest to that residue. Pass both atoms through
    [filt] and only compare distances if it returns true.

    @param prots Array of [protein_chain]s
    @param ligs  A list of [atom] arrays representing the ligands
*)
let protein_ligand_dists filt prots ligs =
  let atom_dist a1 a2 =
    let dx = a1.x -. a2.x in
    let dy = a1.y -. a2.y in
    let dz = a1.z -. a2.z in
    sqrt(dx*.dx +. dy*.dy +. dz*.dz) in
  let mind a1s (d,ma) a2s =			(* Min distance, atom array and list *)
    let dd =ref d in
    let mma = ref ma in
    for i=0 to (Array.length a1s)-1 do
      for j=0 to (Array.length a2s)-1 do
	if filt a1s.(i) a2s.(j) then
	  let dt = atom_dist a1s.(i) a2s.(j) in
	  if dt < !dd then
	    (mma := a2s.(j); dd := dt);
      done;
    done;
    (!dd,!mma) in
  let nprot = Array.length prots in
  let nres = Array.map (fun f -> f.chainlen) prots in
  let dists = Array.init nprot (fun i-> Array.make nres.(i) (infinity,empty_atom)) in
  for i=0 to nprot-1 do
    for j=0 to nres.(i)-1 do
      let mymind = mind prots.(i).atoms.(j) in
      dists.(i).(j) <- List.fold_left mymind dists.(i).(j) ligs;
    done;
  done;
  dists;;

(** Top level command to output the ligand distances for each protein chain *)
let top_prot_lig_dists dist_filt fname pdb =
  let ligfilt ll l = 
    let n = Array.length l in
    if n >= !min_lig_atoms && n <= !max_lig_atoms 
    then l::ll 
    else ll in
  let nucfilt ll na = 
    let l =  na.naallatoms in
    if (Array.length l) >= !min_lig_atoms then l::ll else ll 
  in
  let liglist = 
    (Array.fold_left ligfilt [] pdb.ligands) 
    @ (Array.fold_left nucfilt [] pdb.nuclacids) in
  printf "%s : %d protein chains and %d ligs to consider\n" 
    fname (Array.length pdb.proteins) (List.length liglist);
  let dists = protein_ligand_dists dist_filt pdb.proteins liglist in
  let id = !getcode fname in
  let ext = if !ext = "" then ".dist" else !ext in
  for i=0 to (Array.length dists)-1 do
    let f = open_out (sprintf "%s%s%s" id pdb.proteins.(i).chainid ext) in
    for j=0 to (Array.length dists.(i))-1 do
      let r = pdb.proteins.(i) in
      let d,ma = dists.(i).(j) in
      fprintf f "%8.3f %s %5d %3s %5d %s\n" d r.residues3.(j) r.ids.(j) ma.resName ma.resSeq ma.chainID;
    done;
    close_out f;
  done;;

let print_prot_lig_contacts cutoffs prefix pdb =
  printf "Chain Ligand Contacts\n";
  let ligfilt ll l = if (Array.length l) >= !min_lig_atoms then l::ll else ll in
  let nucfilt ll na = 
    let l =  na.naallatoms in
    if (Array.length l) >= !min_lig_atoms then l::ll else ll 
  in
  let liglist = 
    (Array.fold_left ligfilt [] pdb.ligands) 
    @ (Array.fold_left nucfilt [] pdb.nuclacids) in
  let dists = protein_ligand_dists always_filt pdb.proteins liglist in
  let sumdist cut s (d,_:float*atom) = 
    if d<=cut then s+1 else s in
  for i=0 to (Array.length cutoffs)-1 do
    for j=0 to (Array.length dists)-1 do
      let nc = Array.fold_left (sumdist cutoffs.(i)) 0 dists.(j) in
      printf " %s%s has %d contacts at dist %.2f\n" 
	prefix pdb.proteins.(j).chainid nc cutoffs.(i);
    done;
  done;;

(** Display information on broken chains on stdout *)
let print_broken_info pdb =
  let b = pdb.broken_chains in
  printf "%d Broken Chains\n" (Array.length b);
  for i=0 to (Array.length b)-1 do
    printf " Chain %2s has %2d segments\n" b.(i).(0).alphas.(0).chainID (Array.length b.(i));
    for j=0 to (Array.length b.(i))-1 do
      let n = b.(i).(j).chainlen in
      printf " %2s: %5d residues from %5d to %5d\n" 
	b.(i).(j).chainid n b.(i).(j).ids.(0) b.(i).(j).ids.(n-1);
    done;
  done;;

open Geom
let pymol_bound_box fname pdb =
  let id = !getcode fname in
  let ext = if !ext = "" then ".py" else !ext in
  let f = open_out (id^ext) in
  fprintf f "from pymol.cgo import *\nfrom pymol import cmd\n";
  for i=0 to (Array.length pdb.proteins)-1 do
    let chain = pdb.proteins.(i) in
    let cid = chain.chainid in
    let coords = Util.transpose chain.alpha_coords in
    let cent,v1,_,v3 as bb = Geom.bb_princip_axes coords in
    fprintf f "bb%s%s = [ BEGIN, LINE_STRIP, COLOR, 1.0, 1.0, 1.0,\n" id cid;
    let print v = fprintf f "VERTEX, %f, %f, %f,\n" v.(0) v.(1) v.(2) in
    let vord = 
      let v = Geom.bb_vertices bb in
      [v.(0); v.(1); v.(2); v.(3); v.(0); v.(4); 
       v.(5); v.(1); v.(5); 
       v.(6); v.(2); v.(6); 
       v.(7); v.(3); v.(7);
       v.(4)] 
    in
    List.iter print vord; 
    fprintf f "END ]\ncmd.load_cgo(bb%s%s,'bb%s%s')\n" id cid id cid;
    (* Debugging *)
    if !bbdebug then begin
      let print v = fprintf f "SPHERE, %f, %f, %f, 1.25,\n" v.(0) v.(1) v.(2) in
      fprintf f "bbdbg = [ COLOR, 1.0, 0.0, 0.0,  #1st Axis\n";
      let x = coords in
      print cent;   print (add cent v1);
      fprintf f "COLOR, 0.0, 1.0, 0.0, #2nd Axis\n";
      let op p = let d,_ = line_point_dist (cent,v1) p in d in
      let _,idx = max_over_points op 0.0 x in
      let _,lp = line_point_dist (cent,v1) x.(idx) in
      print lp; print x.(idx);
      let op p = let d,_,_ = plane_point_dist (cent,v3) p in d in
      let _,idx = max_over_points op 0.0 x in
      let _,_,pp = plane_point_dist (cent,v3) x.(idx) in
      fprintf f "COLOR, 0.0, 0.0, 1.0,  #3rd Axis\n";
      print pp; print x.(idx);
      fprintf f "END ]\ncmd.load_cgo(bbdbg,'bbdbg')\n";
    end;
  done;
  close_out f    

let bound_box_str chain =
  let coords = Util.transpose chain.alpha_coords in
  let _,v1,v2,v3 = Geom.bb_princip_axes coords in
  let norm = Geom.norm2 in
  let r1,r2,r3 = (norm v1),(norm v2),(norm v3) in
  sprintf "Axes lens: %6.2f %6.2f %6.2f, Ratios: %4.2f %4.2f %4.2f"
    r1 r2 r3 (r2 /. r1) (r3 /. r1) (r3 /. r2)

(** Print info on the pdb file *)
let pdbinfo fname contact_dists pdb =
  printf "File: %s\n" fname;
(*   printf "%d Entities, chain ids: " (Array.length pdb.ids); *)
(*   Array.iter (fun c-> printf "%s " c) pdb.chains; *)
  printf "%d Protein chains\n" (Array.length pdb.proteins);
  Array.iter (fun r-> printf " %2s length %d residues, %s\n" 
		r.chainid r.chainlen (bound_box_str r))
    pdb.proteins;
  print_broken_info pdb;
  printf "%d NucleicAcid chains\n" (Array.length pdb.nuclacids);
  Array.iter 
    (fun c-> printf " %s with %d acids and %d atoms\n" 
       c.nachainID (Array.length c.acids) (Array.length c.naallatoms))
    pdb.nuclacids;
  printf "%d Ligands\n" (Array.length pdb.ligands);
  Array.iter 
    (fun c-> printf " %s on chain %s with %d atoms\n" 
       c.(0).resName c.(0).chainID (Array.length c)) pdb.ligands;
  printf "%d Other chains\n" (Array.length pdb.others);
  Array.iter (fun c-> printf " %s with %d atoms\n" c.(0).chainID (Array.length c)) pdb.others;
  printf "%d Water molecules\n" (Array.length pdb.water);
  if Array.length contact_dists > 0 then
    let prefix = PDBRead.file_id fname in
    print_prot_lig_contacts contact_dists prefix pdb;
  else ();
  printf "\n";;

(** One line summary of the pdb for creating datasets *)
let pdbsummary fname pdb =
  let code = Util.basename fname in
  let nchains = Array.length pdb.proteins in
  let chain1len = if nchains = 0 then 0 else pdb.proteins.(0).chainlen in
  let hasbroken = if (Array.length pdb.broken_chains) > 0 then 1 else 0 in
  let nligs = Array.length pdb.ligands in
  printf "%8s %3d %4d %1d %3d\n" code nchains chain1len hasbroken nligs


(** Extract the atoms for the flattened residues in fr *)
let extract_residues out fr =
  for i=0 to fr.chainlen-1 do
    for j=0 to (Array.length fr.atoms.(i))-1 do
      let a = fr.atoms.(i).(j) in
      let a = if !renumber then {a with resSeq=i+1} else a in
      let a = if !nohet then {a with hetatom=false} else a in
      let a = if !rechain then {a with chainID=fr.chainid} else a in
      printatom out a;
    done;
  done;
  print_ter out;;

(** Utility to pritn all ligands *)
let extract_ligs out ligs =
  let printatoms atoms = 
    Array.iter (printatom out) atoms; 
    print_ter out in
  Array.iter printatoms ligs;;

(** Extract protein entities into their own file, possibly include the
    ligands based on the module level switches. [lst] is the list of
    entity IDs (chains) to extract *)
let extract_proteins fname pdb lst =
  let ext = if !ext = "" then ".pdb" else !ext in
  let id = !getcode fname in
  let n = Array.length pdb.proteins in
  for i=0 to n-1 do
    if List.mem pdb.proteins.(i).chainid lst then begin
      let newid = id^pdb.proteins.(i).chainid^ext in
      let out = open_out newid in
      extract_residues out pdb.proteins.(i);
      if !include_ligs then extract_ligs out pdb.ligands;
      print_end out;
      close_out out;
    end;
  done;;

let reform_protein fname pdb lst =
  let ext = if !ext = "" then ".2" else !ext in
  let out = open_out (fname^ext) in
  let n = Array.length pdb.proteins in
  for i=0 to n-1 do
    if List.mem pdb.proteins.(i).chainid lst then begin
      extract_residues out pdb.proteins.(i);
    end;
  done;
  if !include_ligs then extract_ligs out pdb.ligands;
  print_end out;
  close_out out;;

(** Print a per chain, per residue description of protein chains in
    the file. Should add ligand and nucleic acid descriptions
    later. *)
let print_hierarch fname pdb =
  let nprots = Array.length pdb.proteins in
  printf "%s with %d chains: " fname nprots;
  for i=0 to nprots-1 do 
    printf "%s " pdb.proteins.(i).chainid;
  done;
  printf "\n";
  for ii=0 to nprots-1 do
    let res = pdb.proteins.(ii) in
    printf " Protein Chain %s with %d residues:\n" res.chainid res.chainlen;
    for i=0 to res.chainlen-1 do
      let atoms = res.atoms.(i) in
      let natoms = (Array.length atoms) in
      printf "  %s %d with %d atoms\n" res.residues3.(i) res.ids.(i) natoms;
      for j=0 to natoms-1 do
	let a = atoms.(j) in
	printf "   %6d %2s %2s %c %2s %4s  %8.3f %8.3f %8.3f\n"
	  a.serial a.namesym a.namebranch a.altLoc a.element a.segID
	  a.x a.y a.z;
      done;
    done;
  done;
  let nligs = Array.length pdb.ligands in
  printf " %d Ligands\n" nligs;
  for ii=0 to nligs-1 do
    let lig = pdb.ligands.(ii) in
    let n = Array.length lig in
    printf "  %s %d on chain %s with %d atoms\n" lig.(0).resName lig.(0).resSeq lig.(0).chainID n;
    for i=0 to n-1 do
      let a = lig.(i) in
      printf "   %6d %2s %2s %c %2s %8.3f %8.3f %8.3f\n"
	a.serial a.namesym a.namebranch a.altLoc a.element a.x a.y a.z;
    done;
  done;
  print_broken_info pdb;;

(** Print dihedral angles for protein *)
let print_angles fname pdb =
  printf "%s with %d chains: " fname (Array.length pdb.proteins);
  let over_chains c = 
    printf " Protein Chain %s with %d residues:\n" c.chainid c.chainlen;
    let over_residues i r3 =
      printf "%4d: %3s %8.2f %8.2f\n" i r3 c.phi.(i) c.psi.(i) 
    in
    Array.iteri over_residues c.residues3
  in
  Array.iter over_chains pdb.proteins

(** Convert a space separated list of characters to a string
    list. Used for the list of chain identifiers. *)
let spsep2list s =
  let n = String.length s in
  let l = ref [] in
  for i=0 to n-1 do
    if s.[i] <> ' ' then 
      let cs = String.make 1 s.[i] in 
      l := cs::!l;
  done;
  List.rev !l;;

(** Open file [fname] and read its content: a whitespace separated of
    files which are returned in a list (reverse order) *)
let read_filelist fname = 
  let b = Scanning.from_file fname in
  let l = ref [] in
  while not (Scanning.end_of_input b) do
    let s = bscanf b " %s " (fun s->s) in
    l := s::!l;
  done;
  !l;;

(* ************************************************** *)
(* MAIN *)
let _ = begin
  parse options deffn usg;
  if !filelist <> "" then		(* Read pdb entries from filelist *)
    flist := List.append (read_filelist !filelist) !flist;
  if !flist = [] then begin		(* Make sure we have something to do *)
    eprintf "%s\n" usg; exit 1
  end;
  flist := List.rev !flist;

  (* Add ligands if present *)
  if !ligfile <> "" then PDBRead.add_ligand_file !ligfile;

  let allchains = !chainlist = "" in
  let deflist = if allchains then [] else spsep2list !chainlist in
  let contact_dists = Array.map float_of_string (Util.parse_strings !contact_dists) in

  (* Read the pdb file *)
  let process fname =
    begin try
      let pdb = read_pdb fname in
      let chain_lst = 
	if allchains then
	  Array.to_list (Array.map (fun p->p.chainid) pdb.proteins) 
	else deflist in
      match !mode with
	"info" -> pdbinfo fname contact_dists pdb
      | "sum"  -> pdbsummary fname pdb
      | "reform" -> reform_protein fname pdb chain_lst
      | "seq" -> pdb2seq fname pdb 
      | "angles" -> print_angles fname pdb
      | "ligdist" -> top_prot_lig_dists !dist_filt fname pdb
      | "extract" -> extract_proteins fname pdb chain_lst
      | "hierarch" -> print_hierarch fname pdb
      | "pybb" -> pymol_bound_box fname pdb
      | m -> eprintf "Bad mode %s : aborting\n" m; exit 1
    with 
      PDBRead.Read_error(_,line) ->
	eprintf "Error with %s : Line %d, unexpected input format, check file\n" fname line
    | Sys_error(s) -> eprintf "Error with %s : %s, skipping\n" fname s
    (* | _ -> eprintf "Bad error with %s, skipping (check this one)\n" fname;(\*  raise e *\) *)
    end;
  in List.iter process !flist;
end;;
